/*
 * This file contains the definition and methods for the
 * JpegSizeRecorder class.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "inet.h"
#include "rtp.h"
#include "rep-decoder.h"
#include "bsd-endian.h"
#include "tclcl.h"
#include "jpeg/jpeg.h"
#include "renderer.h"
#include "pktbuf.h"
#include "vidreps.h"

/* 
 * The JpegSizeRecorder class is used as a measurment tool to 
 * record the size of JPEG frame data. This size includes all
 * RTP and JPEG payload specific headers. The packets are released
 * immediately upon reception and no decoding or other special 
 * handling of the packets is done. The timestamp and size of
 * each JPEG frame is printed to stderr. Partially completed 
 * frames are not reported.
 */

class JpegSizeRecorder : public PacketHandler {
public:
  JpegSizeRecorder() : PacketHandler(sizeof(jpeghdr)), reset_state_(1) {};
  
  virtual void recv(pktbuf *pb);

  int reset_state_;
  int cur_size_;
  u_int32_t last_ts_;
  u_int16_t last_seq_no_;
  int q_;
};

void 
JpegSizeRecorder::recv(pktbuf *pb) 
{
  rtphdr *rh = (rtphdr*)pb->dp;
  jpeghdr *jh = (jpeghdr *) (rh+1);

  int q = jh->q;

  u_int32_t ts = ntohl(rh->rh_ts);
  u_int16_t sn = ntohs(rh->rh_seqno);

  int len = pb->len;

  pb->release();

  if (reset_state_ == 1) {
    cur_size_ = 0;
    last_ts_ = ts;
    last_seq_no_ = sn;
    reset_state_ = 2;
    return;
  } else if (reset_state_ == 2) {
    if (last_ts_ == ts) {
      last_seq_no_ = sn;
      return;
    }
    if (last_seq_no_ != sn - 1) {
      cur_size_ = 0;
      last_ts_ = ts;
      last_seq_no_ = sn;
      reset_state_ = 2;
      return;
    }
    
    cur_size_ = len;
    last_ts_ = ts;
    last_seq_no_ = sn;
    q_ = q;
    reset_state_ = 0;
    return;

  } else {
    if (last_ts_ == ts) {
      if (last_seq_no_ == sn - 1) {
	cur_size_ += len;
	last_seq_no_ = sn;
	return;
      } else {
	cur_size_ = 0;
	last_ts_ = ts;
	last_seq_no_ = sn;
	reset_state_ = 2;
	return;
      }
    } else {
      if (last_seq_no_ == sn - 1) {
	fprintf(stderr, "%u %d %d\n", last_ts_, q_, cur_size_);
	
	cur_size_ = len;
	last_ts_ = ts;
	last_seq_no_ = sn;
	q_ = q;
	reset_state_ = 0;
	return;
      } else {
	cur_size_ = 0;
	last_ts_ = ts;
	last_seq_no_ = sn;
	reset_state_ = 2;
	return;
      }
    }
  }
}	

static class JpegSizeRecorderClass : public TclClass {
public:
  JpegSizeRecorderClass() : TclClass("Module/VideoDecoder/JPEGSizeRecorder") {}
  TclObject *create(int argc, const char*const* argv) {
    return (new JpegSizeRecorder());
  }
} swd_jpeg_size_recorder_;

